package generator

import "fmt"
import "git.tebibyte.media/sashakoshka/fspl/llvm"
import "git.tebibyte.media/sashakoshka/fspl/entity"
import "git.tebibyte.media/sashakoshka/fspl/analyzer"

func (this *generator) generateVariableLoc (variable *entity.Variable) (llvm.Value, error) {
	return this.blockManager.variable(variable.Declaration)
}

func (this *generator) generateSliceLoc (slice *entity.Slice) (llvm.Value, error) {
	var err error
	var start, end, dataAddress llvm.Value
	var sizeType *llvm.TypeInt; {
		ty, err := this.typedef("Index")
		if err != nil { return nil, err }
		sizeType = ty.(*llvm.TypeInt)
	}

	if slice.Start == nil {
		start = llvm.NewConstInt(sizeType, 0)
	} else {
		start, err = this.generateExpression(slice.Start)
		if err != nil { return nil, err }
	}
	if slice.End != nil {
		end, err = this.generateExpression(slice.End)
		if err != nil { return nil, err }
	}

	sourceType := analyzer.ReduceToBase(slice.Slice.Type())
	switch sourceType := sourceType.(type) {
	case *entity.TypeSlice:
		source, err := this.generateExpression(slice.Slice)
		if err != nil { return nil, err }
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }
		dataAddress = this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 0))
		if end == nil {
			endAddr := this.blockManager.NewGetElementPtr (
				irSourceType,
				source,
				llvm.NewConstInt(llvm.I32, 0),
				llvm.NewConstInt(llvm.I32, 2))
			end = this.blockManager.NewLoad(sizeType, endAddr)
		}
		
	case *entity.TypeArray:
		dataAddress, err = this.generateExpressionLoc(slice.Slice)
		if err != nil { return nil, err }
		if end == nil {
			end = llvm.NewConstInt(sizeType, int64(sourceType.Length))
		}
	
	default:
		panic(fmt.Sprint (
			"BUG: generator can't slice expression ",
			slice.Slice))
	}

	
	destinationType, err := this.generateType(slice.Type())
	if err != nil { return nil, err }
	newSlice := this.blockManager.newAllocaFront(destinationType)

	// create new slice descriptor
	dataAddressFieldAddress := this.blockManager.NewGetElementPtr (
		destinationType,
		newSlice,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, 0))
	startFieldAddress := this.blockManager.NewGetElementPtr (
		destinationType,
		newSlice,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, 1))
	endFieldAddress := this.blockManager.NewGetElementPtr (
		destinationType,
		newSlice,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, 2))
	this.blockManager.NewStore(dataAddressFieldAddress, dataAddress)
	this.blockManager.NewStore(startFieldAddress,       start)
	this.blockManager.NewStore(endFieldAddress,         end)

	return newSlice, nil
}

func (this *generator) generateSubscriptLoc (subscript *entity.Subscript) (llvm.Value, error) {
	var sizeType *llvm.TypeInt; {
		ty, err := this.typedef("Index")
		if err != nil { return nil, err }
		sizeType = ty.(*llvm.TypeInt)
	}
	
	source, err := this.generateExpressionLoc(subscript.Slice)
	if err != nil { return nil, err }
	offset, err := this.generateExpression(subscript.Offset)
	if err != nil { return nil, err }

	var elementType entity.Type
	var dataAddress llvm.Value

	sourceType := analyzer.ReduceToBase(subscript.Slice.Type())
	switch sourceType := sourceType.(type) {
	case *entity.TypeSlice:
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }

		// add index to slice start
		startFieldAddress := this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 1))
		start := this.blockManager.NewLoad(sizeType, startFieldAddress)
		offset = this.blockManager.NewAdd(start, offset)

		// get slice element
		elementType = sourceType.Element
		dataAddress = this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 0))
		
	case *entity.TypeArray:
		elementType = sourceType.Element
		dataAddress = source
	
	default:
		panic(fmt.Sprint (
			"BUG: generator can't subscript expression ",
			subscript.Slice))
	}

	irElementType, err := this.generateType(elementType)
	if err != nil { return nil, err }
	return this.blockManager.NewGetElementPtr (
		irElementType,
		dataAddress,
		offset), nil
}

func (this *generator) generateDereferenceLoc (dereference *entity.Dereference) (llvm.Value, error) {
	return this.generateExpression(dereference.Pointer)
}

func (this *generator) generateMemberAccessLoc (access *entity.MemberAccess) (llvm.Value, error) {
	source, err := this.generateExpressionLoc(access.Source)
	if err != nil { return nil, err }
	sourceType := analyzer.ReduceToBase(access.Source.Type()).(*entity.TypeStruct)
	irSourceType, err := this.generateType(sourceType)
	if err != nil { return nil, err }

	offset := -1
	for index, name := range sourceType.MemberOrder {
		if name == access.Member {
			offset = index
			break
		}
	}
	
	return this.blockManager.NewGetElementPtr (
		irSourceType,
		source,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, int64(offset))), nil
}

func (this *generator) generateBlockLoc (block *entity.Block) (llvm.Value, error) {
	var last llvm.Value
	var err error
	for index, step := range block.Steps {
		// prefer generating a normal statement to avoid allocations
		if index == len(block.Steps) - 1 {
			last, err = this.generateStatementLoc(step)
		} else {
			last, err = this.generateStatement(step)
		}
		if err != nil { return nil, err }
	}
	return last, nil
}

